"""
A large portion of this code is adapted from MQTT Connect for Meshtastic, version 0.8.7
https://github.com/pdxlocations/connect/blob/869be93a3c32d9550cbd63c1ae0ccb61686eca60/mqtt-connect.py
A number of functions have been broken up and reorganized to make it more usable as a library.
"""

import base64
import os
from configparser import ConfigParser
from contextlib import suppress
from typing import Any, Dict, TypedDict


def xor_checksum(data: bytes) -> int:
    checksum: int = 0
    for char in data:
        checksum ^= char
    return checksum


class ConfigSet(TypedDict):
    host: str
    port: int
    username: str
    password: str
    root_topic: str
    channel: str
    userid: int
    key: str


class MQTTConfig:
    def __init__(self):
        self.defaults()
        self.load_config()
        self.load_env()

        # Replace the default shortened key with the full key
        if self.config["key"] == "AQ==":
            self.config["key"] = "1PG7OiApB1nwvP+rz05pAQ=="

    def defaults(self):
        self.config: ConfigSet = {
            "host": "mqtt.meshtastic.org",
            "port": 1883,
            "username": "meshdev",
            "password": "large4cats",
            "root_topic": "msh/US",
            "channel": "LongFast",
            "userid": 452664778,
            "key": "AQ==",
        }

    def load_config(self):
        config = ConfigParser()
        with suppress(FileNotFoundError):
            config.read("mqtt.conf")
            for key in self.config:
                value = config.get("mqtt", key)
                if isinstance(self.config[key], int):
                    value = int(value)
                if value:
                    self.config[key] = value

    def load_env(self):
        for key in self.config:
            value = os.getenv(f"MQTT_{key.upper()}")
            if value:
                self.config[key] = value

    @property
    def userid(self) -> str:
        return "!%08x" % self.config["userid"]

    @property
    def topic(self) -> str:
        return f"{self.config['root_topic']}/2/e/{self.config['channel']}/{self.userid}"

    @property
    def key(self) -> bytes:
        return base64.b64decode(self.config["key"].encode("ascii"))

    @property
    def encoded_channel(self) -> int:
        key = self.config["key"].replace("-", "+").replace("_", "/")
        key_bytes = base64.b64decode(key.encode("utf-8"))
        h_key = xor_checksum(key_bytes)
        h_name = xor_checksum(self.config["channel"].encode("utf-8"))
        return h_name ^ h_key

    @property
    def aiomqtt_config(self) -> Dict[str, str | int]:
        return {
            "hostname": self.config["host"],
            "port": self.config["port"],
            "username": self.config["username"],
            "password": self.config["password"],
        }



from meshtastic.protobuf import portnums_pb2

from .message import MeshtasticMessage
from .config import MQTTConfig


class MeshtasticTextMessage(MeshtasticMessage):
    def __init__(self, payload: str, config: MQTTConfig):
        self.type = portnums_pb2.TEXT_MESSAGE_APP
        super().__init__(payload.encode("utf-8"), config)



from meshtastic.protobuf import portnums_pb2, mesh_pb2

from .message import MeshtasticMessage
from .config import MQTTConfig


class MeshtasticNodeInfoMessage(MeshtasticMessage):
    def __init__(self, config: MQTTConfig):
        self.type = portnums_pb2.NODEINFO_APP
        payload = mesh_pb2.User()
        payload.id = config.userid
        payload.short_name = "MQTT"
        payload.long_name = f"Meshage {config.userid}"
        payload.hw_model = mesh_pb2.HardwareModel.PRIVATE_HW
        payload.is_unmessagable = True

        super().__init__(payload.SerializeToString(), config)



"""
An object oriented interface for sending and receiving Meshtastic messages via MQTT.

A large portion of this code is adapted from MQTT Connect for Meshtastic, version 0.8.7
https://github.com/pdxlocations/connect/blob/869be93a3c32d9550cbd63c1ae0ccb61686eca60/mqtt-connect.py
A number of functions have been broken up and reorganized to make it more usable as a library.
"""

import random
from abc import ABC, abstractmethod

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from meshtastic import BROADCAST_NUM
from meshtastic.protobuf import mesh_pb2, mqtt_pb2, portnums_pb2

from .config import MQTTConfig


class MeshtasticMessage(ABC):
    type: portnums_pb2.PortNum.ValueType

    @abstractmethod
    def __init__(self, payload: bytes, config: MQTTConfig):
        self.config = config
        self.message_id = self.generate_message_id()
        self.payload = payload

    def generate_message_id(self) -> int:
        return random.getrandbits(32)

    def encrypt_packet(self, packet: mesh_pb2.MeshPacket) -> bytes:
        # Wrap payload in Data protobuf message
        data_msg = mesh_pb2.Data()
        data_msg.portnum = self.type
        data_msg.payload = self.payload
        data_msg.bitfield = 1

        nonce_packet_id = packet.id.to_bytes(8, "little")
        nonce_from_node = getattr(packet, "from").to_bytes(8, "little")
        nonce = nonce_packet_id + nonce_from_node
        encryptor = Cipher(
            algorithms.AES(self.config.key), modes.CTR(nonce), backend=default_backend()
        ).encryptor()
        return encryptor.update(data_msg.SerializeToString()) + encryptor.finalize()

    def packet(self) -> mesh_pb2.MeshPacket:
        packet = mesh_pb2.MeshPacket()
        packet.id = self.message_id
        setattr(
            packet, "from", self.config.config["userid"]
        )  # from is a reserved keyword
        packet.to = BROADCAST_NUM
        packet.want_ack = False
        packet.channel = self.config.encoded_channel
        packet.hop_limit = 3
        packet.hop_start = 3
        packet.encrypted = self.encrypt_packet(packet)
        return packet

    def service_envelope(self, packet: mesh_pb2.MeshPacket) -> mqtt_pb2.ServiceEnvelope:
        envelope = mqtt_pb2.ServiceEnvelope()
        envelope.packet.CopyFrom(packet)
        envelope.channel_id = self.config.config["channel"]
        envelope.gateway_id = self.config.userid
        return envelope

    def __bytes__(self) -> bytes:
        return self.service_envelope(self.packet()).SerializeToString()
